using System;
using System.Collections.Generic;
using System.Linq;
using MacGruber;
using MeshVR;
using MVR.FileManagementSecure;
using SimpleJSON;
using UnityEngine;
using UnityEngine.UI;


namespace CheesyFX
{
    public class ShakeIt : MVRScript
    {
        public static string morphPath = "Custom/Atom/Person/Morphs/female/CheesyFX/ShakeIt/";
        public static string flexMorphPath = "Custom/Atom/Person/Morphs/female/CheesyFX/FlexIt/";
        private const string saveDir = "Saves/PluginData/CheesyFX/ShakeIt/";
        private static JSONClass factoryDefaults;

        public static DAZCharacterSelector.Gender gender;
        private DAZCharacterSelector dcs;

        private List<JiggleHandler> selectedHandlers;
        public static int selectedRegionId;

        private static string[][] rbNames1 = 
        {
            new[] {"Glutes", "LGlute", "RGlute" },
            new[] {"Thigh", "lThigh", "rThigh" },
            new[] {"Calf",  "lShin", "rShin" },
            new[] {"Abs",  "abdomen" },
            new[]{"AbsSide", "abdomen"},
            new[] {"UArm",  "lShldr", "rShldr" },
            new[] {"Breasts",  "lPectoral", "rPectoral" },
            new[] {"Cheeks",  "head"},
        };

        private static List<JiggleRegion> jiggleRegions = new List<JiggleRegion>();

        public static Dictionary<int, float> sin = new Dictionary<int, float>();

        private UIDynamicV4Slider maxMorph;
        private UIDynamicV4Slider strengthUid;
        private UIDynamicV4Slider frequencyUid;
        private UIDynamicV4Slider dampingUid;
        private UIDynamicV4Slider minAccelUid;
        private UIDynamicV4Slider maxAccelUid;
        private UIDynamicV4Slider upperUid;
        private UIDynamicV4Slider midUid;
        private UIDynamicV4Slider lowerUid;
        
        private UIDynamicTabBar regionTabbar;
        private UIDynamicTabBar subTabbar;
        private UIDynamicToggleArray toggleAray;
        public static UIDynamicToggle monitorUid;

        private static GenerateDAZMorphsControlUI morphUI;
        public JSONStorableUrl loadURL;
        private bool needsStore;

        private static bool develop = false;

        private JSONStorableBool combined = new JSONStorableBool("Combined", false);
        private JSONStorableBool regionEnabled = new JSONStorableBool("Region Enabled", true);
        private JSONStorableBool tensionEnabled = new JSONStorableBool("Tension Enabled", false);
        // private JSONStorableBool useCustomMorphs = new JSONStorableBool("Custom Morphs", false);
        private JSONStorableBool configureTensedState = new JSONStorableBool("Configure tensed state", false);
        private JSONStorableBool monitorJ = new JSONStorableBool("Show tension", false);

        public static bool monitor;

        // private MyJSONStorableVector3 vec = new MyJSONStorableVector3("vec",new Vector3(.5f,.5f,.5f),
        //     new Vector3(), new Vector3(1f,1f,1f));

        public static JSONStorableString infoLeft = new JSONStorableString("Info", "");
        private UnityEventsListener _uiListener;

        public override void Init()
        {
            if (develop)
            {
                SuperController.singleton.ClearMessages();
                SuperController.singleton.ClearErrors();
            }
            dcs = containingAtom.GetComponentInChildren<DAZCharacterSelector>();
            gender = dcs.gender;
            if (gender == DAZCharacterSelector.Gender.Male)
            {
                morphPath = morphPath.Replace("female", "male");
                flexMorphPath = flexMorphPath.Replace("female", "male");
                rbNames1[0] = new[] { "Glutes", "lThigh", "rThigh" };
            }
            morphUI = dcs.morphsControlUI;
            FileManagerSecure.CreateDirectory(saveDir);

            if (Utils.IsInPackage(this))
            {
                var packageUid = Utils.GetPluginPath(this).Split(':')[0] + ":/";
                morphPath = packageUid + morphPath;
                flexMorphPath = packageUid + flexMorphPath;
            }
            
            loadURL = new JSONStorableUrl("loadURL", "", UILoadJSON, "json", true);
            loadURL.hideExtension = true;
            loadURL.setCallbackFunction -= UILoadJSON;
            loadURL.allowFullComputerBrowse = false;
            loadURL.allowBrowseAboveSuggestedPath = true;
            loadURL.SetFilePath(saveDir);
            loadURL.setCallbackFunction += UILoadJSON;

            for (int i=0; i<rbNames1.Length; i++)
            {
                JiggleRegion jr = gameObject.AddComponent<JiggleRegion>();
                jr.Init(this, i, rbNames1[i][0], rbNames1[i].Skip(1).ToArray());
                jiggleRegions.Add(jr);
            }

            
            const float twoPi = 2f * Mathf.PI;
            for (int i = 0; i < 360; i++)
            {
                sin[i] = Mathf.Sin(i/360f*twoPi);
            }

            regionEnabled.setCallbackFunction += OnRegionEnabled;
            combined.setCallbackFunction += val =>
            {
                subTabbar.ToggleButtons(!val);
                jiggleRegions[regionTabbar.id].isCombined.val = val;
                if(!val) subTabbar.SelectTab(jiggleRegions[regionTabbar.id].selectedSubregion);
                else RegisterHandlers();
            };
            tensionEnabled.setCallbackFunction += val =>
            {
                jiggleRegions[regionTabbar.id].tensionEnabled.val = val;
            };
            configureTensedState.setCallbackFunction += val =>
            {
                RegisterHandlers();
            };
            monitorJ.setCallbackFunction += val =>
            {
                monitor = val;
                strengthUid.SetInteractable(!val);
                frequencyUid.SetInteractable(!val);
                dampingUid.SetInteractable(!val);
                if (val)
                {
                    strengthUid.RegisterVector(selectedHandlers[0].strengthMonitorJSON);
                    frequencyUid.RegisterVector(selectedHandlers[0].frequencyMonitorJSON);
                    dampingUid.RegisterVector(selectedHandlers[0].dampingMonitorJSON);
                    
                }
                else
                {
                    monitorUid.label = monitorJ.name;
                    RegisterHandlers();
                }
            };

            UIManager.OnInitUI(CreateUIElement, leftUIElements, rightUIElements);
            Utils.OnInitUI(CreateUIElement);
            CreateUI();
            factoryDefaults = MyGetJSON();
            ApplyLatestMatchingPreset();
        }
        
        public override void InitUI()
        {
            base.InitUI();
            if(UITransform == null || _uiListener != null)
            {
                return;
            }

            _uiListener = UITransform.gameObject.AddComponent<UnityEventsListener>();
            if(_uiListener != null)
            {
                _uiListener.onEnabled.AddListener(() =>
                {
                    monitorJ.val = false;
                });
            }
        }

        public void Update()
        {
            if (dcs.gender != gender)
            {
                containingAtom.GetComponentInChildren<MVRPluginManager>().ReloadPluginWithUID(name.Substring(0, name.IndexOf('_')));
            }
        }

        public void OnEnable()
        {
            jiggleRegions.ForEach(x => x.enabled = x.enabledJ.val);
            regionEnabled.toggle.interactable = true;
        }
        
        public void OnDisable()
        {
            jiggleRegions.ForEach(x => x.enabled = false);
            regionEnabled.toggle.interactable = false;
        }

        private void OnDestroy()
        {
            jiggleRegions.ForEach(Destroy);
            // Destroy(debugObject);
            if(_uiListener != null)
            {
                DestroyImmediate(_uiListener);
            }
        }

        private JSONStorableStringChooser[] morphChoosers = new JSONStorableStringChooser[]
        {
            new JSONStorableStringChooser("MorphX", null, "", "MorphX"),
            new JSONStorableStringChooser("MorphY", null, "", "MorphY"),
            // new JSONStorableStringChooser("MorphZ", null, "", "MorphZ"),
            // new JSONStorableStringChooser("MorphW", null, "", "MorphW")
        };

        private JSONStorableStringChooser morphChooser =
            new JSONStorableStringChooser("morphChooser", null, "", "Morph");
        
        private void CreateUI()
        {
            var twinButton = Utils.SetupTwinButton(this,
                "Save preset", UISaveJSONDialog,
                "Load preset", delegate {}, false);
            loadURL.RegisterFileBrowseButton(twinButton.buttonRight);
            Utils.SetupTwinButton(this,
                "UserDefaults", () => ApplyFromUrl(saveDir+"UserDefaults.json"),
                "FactoryDefaults", () => ApplyJSON(factoryDefaults), true);
            string[] regions =
            {
                "Glutes", "Thighs", "Calves", "Abs", "SideAbs", "UArms", "Breasts", "Cheeks"
            };
            regionTabbar = UIManager.CreateTabBar(regions, OnRegionChanged, script:this, invokeDefault:false);
            toggleAray = UIManager.CreateToggleArray(new[] { regionEnabled, combined, tensionEnabled }, script:this);
            // toggleAray.labels.ForEach(x => x.fontSize = 24);
            subTabbar = UIManager.CreateTabBar(new []{"Upper", "Mid", "Lower"}, OnSubRegionChanged, script:this, itemHeight:45f);
            // UIManager.CreateToggleArray(new[] { configureTensedState, monitor }, script:this);
            var uid = CreateToggle(configureTensedState);
            uid.ForceHeight(35f);
            monitorUid = CreateToggle(monitorJ, true);
            monitorUid.ForceHeight(35f);
            GetSelectedHandlers();
            CreateSliders();

            // UIManager.CreateChooserArray(morphChoosers, script: this);
            regionTabbar.SelectTab(0);
            subTabbar.SelectTab(0);
        }

        private void OnRegionEnabled(bool val)
        {
            jiggleRegions[regionTabbar.id].enabledJ.val = val;
        }

        private void OnRegionChanged(int id)
        {
            selectedRegionId = id;
            configureTensedState.val = false;
            var region = jiggleRegions[id];
            if (!region.isCombined.val && !combined.val) subTabbar.SelectTab(region.selectedSubregion);
            else RegisterHandlers();
            combined.val = region.isCombined.val;
            regionEnabled.valNoCallback = region.enabledJ.val;
            tensionEnabled.valNoCallback = region.tensionEnabled.val;
            toggleAray.toggles[2].gameObject.SetActive(id < 6);
        }

        private void OnSubRegionChanged(int id)
        {
            jiggleRegions[regionTabbar.id].selectedSubregion = id;
            RegisterHandlers();
        }

        private void GetSelectedHandlers()
        {
            var region = jiggleRegions[regionTabbar.id];
            int subId = subTabbar.id;
            var handlers = region.handlers;
            if (region.isCombined.val) selectedHandlers = handlers;
            else
            {
                if (handlers.Count == 3) selectedHandlers = new List<JiggleHandler> { handlers[subId] };
                else
                {
                    selectedHandlers = new List<JiggleHandler>
                    {
                        handlers[subId],
                        handlers[subId + 3]
                    };
                }
            }
        }

        private void RegisterHandlers(bool bent = false)
        {
            if (monitorJ.val) monitorJ.val = false;
            GetSelectedHandlers();
            if(!configureTensedState.val)
            {
                strengthUid.RegisterVectors(selectedHandlers.Select(x => x.strengthJSON).ToList());
                frequencyUid.RegisterVectors(selectedHandlers.Select(x => x.frequencyJSON).ToList());
                dampingUid.RegisterVectors(selectedHandlers.Select(x => x.dampingJSON).ToList());
                minAccelUid.RegisterVectors(selectedHandlers.Select(x => x.minAccelJSON).ToList());
                maxAccelUid.RegisterVectors(selectedHandlers.Select(x => x.maxAccelJSON).ToList());
                maxMorph.RegisterVectors(selectedHandlers.Select(x => x.maxMorphJSON).ToList());
            }
            else
            {
                strengthUid.RegisterVectors(selectedHandlers.Select(x => x.strengthTenseJSON).ToList());
                frequencyUid.RegisterVectors(selectedHandlers.Select(x => x.frequencyTenseJSON).ToList());
                dampingUid.RegisterVectors(selectedHandlers.Select(x => x.dampingTenseJSON).ToList());
                minAccelUid.RegisterVectors(selectedHandlers.Select(x => x.minAccelJSON).ToList());
                maxAccelUid.RegisterVectors(selectedHandlers.Select(x => x.maxAccelJSON).ToList());
                maxMorph.RegisterVectors(selectedHandlers.Select(x => x.maxMorphJSON).ToList());
            }
            for (int i = 0; i < 4; i++)
            {
                bool active = selectedHandlers[0].morphs[i] != null;
                strengthUid.SliderSetActive(i, active);
                frequencyUid.SliderSetActive(i, active);
                dampingUid.SliderSetActive(i, active);
                minAccelUid.SliderSetActive(i, active);
                maxAccelUid.SliderSetActive(i, active);
                maxMorph.SliderSetActive(i, active);
            }
        }

        private void CreateSliders()
        {
            strengthUid = UIManager.CreateV4Slider(script: this);
            frequencyUid = UIManager.CreateV4Slider(script: this);
            dampingUid = UIManager.CreateV4Slider(script: this);
            minAccelUid = UIManager.CreateV4Slider(script: this);
            maxAccelUid = UIManager.CreateV4Slider(script: this);
            maxMorph = UIManager.CreateV4Slider(script: this);

            strengthUid.setToggleCallbackFunction += strengthUid.ToggleSync;
            strengthUid.toggle.label = "Sync";
            frequencyUid.setToggleCallbackFunction += frequencyUid.ToggleSync;
            frequencyUid.toggle.label = "Sync";
            dampingUid.setToggleCallbackFunction += dampingUid.ToggleSync;
            dampingUid.toggle.label = "Sync";
            minAccelUid.setToggleCallbackFunction += minAccelUid.ToggleSync;
            minAccelUid.toggle.label = "Sync";
            maxAccelUid.setToggleCallbackFunction += maxAccelUid.ToggleSync;
            maxAccelUid.toggle.label = "Sync";
            maxMorph.setToggleCallbackFunction += maxMorph.ToggleSync;
            maxMorph.toggle.label = "Sync";
            regionTabbar.SelectLast();
        }
        
        // GameObject debugObject;
        // public void Debug(Transform parent = null)
        // {
        //     debugObject = GameObject.CreatePrimitive(PrimitiveType.Sphere);
        //     debugObject.transform.localScale = new Vector3(.005f, .005f, .005f);
        //     Destroy(debugObject.GetComponent<Collider>());
        //     var renderer = debugObject.GetComponent<Renderer>();
        //     renderer.material.shader = Shader.Find("Battlehub/RTGizmos/Handles");
        //     renderer.material.color = new Color(0f, 0.89f, 1f);
        //     renderer.enabled = true;
        //     if (parent)
        //     {
        //         debugObject.transform.SetParent(parent);
        //         debugObject.transform.localPosition = new Vector3();
        //         debugObject.transform.localRotation = new Quaternion();
        //     }
        //     var lr = debugObject.AddComponent<LineRenderer>();
        //     // lr.useWorldSpace = false;
        //     lr.SetWidth(0.002f, 0.002f);
        //     lr.SetPositions(new Vector3[]{new Vector3(), debugObject.transform.right*5f});
        // }
        
        public static void SetupTension(bool flexEnabled)
        {
            jiggleRegions.ForEach(x => x.SetupTension(flexEnabled));
        }
        
        public static DAZMorph GetMorphByUid(string name, string directory = null)
        {
            if (directory == null) directory = morphPath;
            DAZMorph morph;
            string uid = directory + name +".vmi";
            morph = morphUI.GetMorphByUid(uid);
            // if(morph == null) $"Morph '{uid}' not found!".Print();
            return morph;
        }
        
        public JSONClass MyGetJSON()
        {
            JSONClass jc = new JSONClass();
            for (int i = 0; i < jiggleRegions.Count; i++)
            {
                jiggleRegions[i].Reset();
                jiggleRegions[i].Store(jc, true);
            }
            return jc;
        }
        
        private void ApplyFromUrl(string path)
        {
            if(!FileManagerSecure.FileExists(path)) return;
            path = FileManagerSecure.NormalizePath(path);
            {
                JSONClass jc = LoadJSON(path).AsObject;
                ApplyJSON(jc);
            }
        }

        public void ApplyJSON(JSONClass jc)
        {
            for (int i = 0; i < jiggleRegions.Count; i++)
            {
                jiggleRegions[i].Load(jc);
            }
            regionTabbar.SelectLast();
        }
        
        public void UISaveJSONDialog(){
            SuperController sc = SuperController.singleton;
            sc.GetMediaPathDialog(UISaveJSON, "json", saveDir, false, true, false, null, true, null, false, false);
            sc.mediaFileBrowserUI.SetTextEntry(true);
            if(sc.mediaFileBrowserUI.fileEntryField != null){
                string filename = GetSuggestedPresetName(containingAtom.name, saveDir);
                sc.mediaFileBrowserUI.fileEntryField.text = filename;
                sc.mediaFileBrowserUI.ActivateFileNameField();
            }
        }
        
        public void UILoadJSON(string url){
            JSONClass jc = LoadJSON(url).AsObject;
            if(jc != null) ApplyJSON(jc);
        }
        
        public void UISaveJSON(string path)
        {
            if (string.IsNullOrEmpty(path)) return;
            string path1 = path.Replace('\\', '/')+".json";
            JSONClass jc = MyGetJSON();
            for (int i=0; i<jiggleRegions.Count; i++) jiggleRegions[i].Reset();
            SaveJSON(jc, path1);
            SuperController.singleton.DoSaveScreenshot(path+".jpg");
        }

        private void ApplyLatestMatchingPreset()
        {
            List<string> matchingPresetPaths = new List<string>();
            string atomName = containingAtom.name;
            string[] paths = FileManagerSecure.GetFiles(saveDir);
            for (int i = 0; i < paths.Length; i++)
            {
                string path = paths[i];
                if (!path.EndsWith(".json")) continue;
                string presetName = path.Replace(saveDir+"\\","");
                
                if(presetName.StartsWith(atomName)) matchingPresetPaths.Add(presetName);
                else
                {
                    if(gender == DAZCharacterSelector.Gender.Male && presetName.StartsWith("UserDefaultsMale")) matchingPresetPaths.Insert(0, presetName);
                    else if (presetName.StartsWith("UserDefaults")) matchingPresetPaths.Insert(0, presetName);
                }
            }

            if (matchingPresetPaths.Count > 0)
            {
                string prestName = matchingPresetPaths.Last();
                ApplyFromUrl(saveDir+prestName);
                $"Successfully applied '{saveDir+prestName}'.".Print();
            }
        }

        public static string GetSuggestedPresetName(string atomName, string directory)
        {
            List<string> presetNames = new List<string>();
            string[] paths = FileManagerSecure.GetFiles(directory);
            for (int i = 0; i < paths.Length; i++)
            {
                string path = paths[i];
                if (!path.EndsWith(".json")) continue;
                string preset = path.Replace(directory+"\\","");
                preset = preset.Remove(preset.Length - 3);
                presetNames.Add(preset);
            }
            string suggested;
            for (int i = 1; i < 20; i++)
            {
                suggested = atomName+ "." + i.ToString("00");
                if (!presetNames.Contains(suggested)) return suggested;
            }
            return atomName;
        }
        
        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            JSONClass tempJC = MyGetJSON();
            JSONClass jc = base.GetJSON(includePhysical, includeAppearance, true);
            jc["ShakeIt"] = tempJC;
            return jc;
        }
        
        public override void LateRestoreFromJSON(
            JSONClass jc,
            bool restorePhysical,
            bool restoreAppearance,
            bool setMissingToDefault
        )
        {
            base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
            if (restorePhysical && restoreAppearance) // StoreType.Full
            {
                if (jc.HasKey("ShakeIt")){
                    JSONClass jcShakeIt = jc["ShakeIt"].AsObject;
                    ApplyJSON(jcShakeIt);
                }
                
            }
        }
    }
}