using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using MeshVR;
using MVR.FileManagementSecure;
using UnityEngine;
using UnityEngine.UI;

namespace CheesyFX
{
    public class ShakeIt_SessionPlugin : MVRScript
    {
        private static string shakeItUrl;
        private static string heelAdjustUrl;
        private JSONStorableUrl femalePresetUrl;
        private JSONStorableUrl malePresetUrl;

        private JSONStorableStringChooser loadShakeItOn = new JSONStorableStringChooser("loadShakeItOn",
            new List<string> { "All", "Females", "Males", "Nobody"}, "All", "Load On");
        private JSONStorableStringChooser loadHeelAdjustOn = new JSONStorableStringChooser("loadHeelAdjustOn",
            new List<string> { "All", "Females", "Males", "Nobody"}, "Females", "Load On");
        private JSONStorableStringChooser loadPresetOn = new JSONStorableStringChooser("loadPresetOn",
            new List<string> { "All", "Females", "Males", "Nobody"}, "Females", "Load Preset On");

        private JSONStorableBool shakeItLoadEnabled = new JSONStorableBool("Load Enabled (SI)", true);
        private JSONStorableBool heelAdjustLoadEnabled = new JSONStorableBool("Load Enabled (HA)", true);

        private void OnEnable()
        {
            SuperController.singleton.onAtomAddedHandlers += AddPerson;
        }

        public override void Init()
        {
            string id = name.Substring(0, name.IndexOf('_'));
            shakeItUrl = manager.GetJSON()["plugins"][id].Value.Replace("_SessionPlugin.cs", ".cslist");
            heelAdjustUrl = "Custom/Scripts/CheesyFX/HeelAdjust2/HeelAdjust2.cslist";
            if (shakeItUrl.StartsWith("CheesyFX."))
            {
                string packageUid = GetLatestPackageUid("CheesyFX.HeelAdjust2");
                if(packageUid != null) heelAdjustUrl = packageUid + ":/" + heelAdjustUrl;
                else
                {
                    loadHeelAdjustOn.val = "HeelAdjust2 missing";
                    loadHeelAdjustOn.choices = new List<string> { "HeelAdjust2 missing" };
                }
            }
            
            femalePresetUrl = new JSONStorableUrl("presetUrl", "", "vap", "Custom/Atom/Person/Plugins");
            femalePresetUrl.hideExtension = true;
            femalePresetUrl.allowFullComputerBrowse = false;
            femalePresetUrl.allowBrowseAboveSuggestedPath = true;
            
            malePresetUrl = new JSONStorableUrl("presetUrl", "", "vap", "Custom/Atom/Person/Plugins");
            malePresetUrl.hideExtension = true;
            malePresetUrl.allowFullComputerBrowse = false;
            malePresetUrl.allowBrowseAboveSuggestedPath = true;
            
            RegisterStringChooser(loadShakeItOn);
            RegisterStringChooser(loadHeelAdjustOn);
            RegisterStringChooser(loadPresetOn);
            RegisterUrl(femalePresetUrl);
            RegisterUrl(malePresetUrl);
            RegisterBool(shakeItLoadEnabled);
            RegisterBool(heelAdjustLoadEnabled);
            
            RegisterPersons();
            CreateUI();
        }

        public void CreateUI()
        {
            var textfield = CreateTextField(new JSONStorableString("bla", "<b>ShakeIt</b>"));
            textfield.GetComponentInChildren<Text>().alignment = TextAnchor.MiddleCenter;
            LayoutElement component = textfield.GetComponent<LayoutElement>();
            component.minHeight = 35f;
            component.preferredHeight = 35f;
            CreateToggle(shakeItLoadEnabled);
            CreatePopup(loadShakeItOn).popupPanelHeight = 275f;
            
            
            textfield = CreateTextField(new JSONStorableString("bla", "<b>HeelAdjust2</b>"), true);
            textfield.GetComponentInChildren<Text>().alignment = TextAnchor.MiddleCenter;
            component = textfield.GetComponent<LayoutElement>();
            component.minHeight = 35f;
            component.preferredHeight = 35f;
            CreateToggle(heelAdjustLoadEnabled,true);
            CreatePopup(loadHeelAdjustOn, true).popupPanelHeight = 275f;


            textfield = CreateTextField(new JSONStorableString("blub", "Additional Plugin Presets"));
            textfield.GetComponentInChildren<Text>().alignment = TextAnchor.MiddleCenter;
            component = textfield.GetComponent<LayoutElement>();
            component.minHeight = 35f;
            component.preferredHeight = 35f;
            var spacer = CreateSpacer(true);
            spacer.height = 35f;
            var button = CreateButton("Choose Female Preset");
            femalePresetUrl.RegisterFileBrowseButton(button.button);
            CreateTextField(femalePresetUrl);
            CreateButton("Reset").button.onClick.AddListener(() => femalePresetUrl.val = "");
            button = CreateButton("Choose Male Preset", true);
            malePresetUrl.RegisterFileBrowseButton(button.button);
            CreateTextField(malePresetUrl, true);
            CreateButton("Reset", true).button.onClick.AddListener(() => malePresetUrl.val = "");
        }
        
        public void OnDisable()
        {
            SuperController.singleton.onAtomAddedHandlers -= AddPerson;
        }

        public void OnDestroy()
        {
            SuperController.singleton.onAtomAddedHandlers -= AddPerson;
        }

        private void RegisterPersons()
        {
            foreach (var atom in SuperController.singleton.GetAtoms())
            {
                 AddPerson(atom);
            }
        }

        private void AddPerson(Atom atom)
        {
            if (atom.type != "Person") return;
            StartCoroutine(AddPlugins(atom));
        }

        private IEnumerator AddPlugins(Atom person)
        {
            while (SuperController.singleton.isLoading) yield return new WaitForSeconds(.5f);
            yield return new WaitForEndOfFrame();
            
            var dcs = person.GetComponentInChildren<DAZCharacterSelector>();
            while ((object)dcs == null)
            {
                yield return new WaitForSeconds(.5f);
                dcs = person.GetComponentInChildren<DAZCharacterSelector>();
            }

            while (!dcs.selectedCharacter.ready)
            {
                yield return new WaitForSeconds(.5f);
            }
            var pm = person.GetComponentInChildren<MVRPluginManager>();
            var plugins = pm.GetComponentsInChildren<MVRScript>();
            var hasShakeIt = plugins.Any(x => x.name.EndsWith("CheesyFX.ShakeIt"));
            var hasHeelAdjust = plugins.Any(x => x.name.EndsWith("CheesyFX.HeelAdjust2"));
            var gender = dcs.gender;
            var loadShakeIt = !hasShakeIt && (loadShakeItOn.val == "All" ||
                              (gender == DAZCharacterSelector.Gender.Female && loadShakeItOn.val == "Females") ||
                              (gender == DAZCharacterSelector.Gender.Male && loadShakeItOn.val == "Males"));
            var loadHeelAdjust = !hasHeelAdjust && (loadHeelAdjustOn.val == "All" ||
                              (gender == DAZCharacterSelector.Gender.Female && loadHeelAdjustOn.val == "Females") ||
                              (gender == DAZCharacterSelector.Gender.Male && loadHeelAdjustOn.val == "Males"));
            if(loadShakeIt)
            {
                var plugin = pm.CreatePlugin();
                plugin.pluginURLJSON.val = shakeItUrl;
                if (!shakeItLoadEnabled.val) plugin.scriptControllers.ForEach(x => x.script.enabledJSON.val = false);
            }
            if (loadHeelAdjust)
            {
                MVRPlugin plugin = pm.CreatePlugin();
                plugin.pluginURLJSON.val = heelAdjustUrl;
                if (!heelAdjustLoadEnabled.val) plugin.scriptControllers.ForEach(x => x.script.enabledJSON.val = false);
            }
            var pluginPresetManager = person.GetStorableByID("PluginPresets").GetComponentInChildren<PresetManager>();
            var pluginPresetManagerControl = person.presetManagerControls.Find(pmc => pmc.name == "PluginPresets");
            if (gender == DAZCharacterSelector.Gender.Female && femalePresetUrl.val != "")
            {
                pluginPresetManagerControl.SetStringParamValue("presetName", femalePresetUrl.val.Substring(34).Replace(".vap", ""));
                pluginPresetManager.MergeLoadPreset();
            }
            else if (gender == DAZCharacterSelector.Gender.Male && malePresetUrl.val != "")
            {
                pluginPresetManagerControl.SetStringParamValue("presetName", malePresetUrl.val.Substring(34).Replace(".vap", ""));
                pluginPresetManager.MergeLoadPreset();
            }
        }
        
        private static string GetLatestPackageUid(string packageUid){
            try
            {
                int version = FileManagerSecure.GetPackageVersion(packageUid+".latest");
                if (version == -1) return null;
                return $"{packageUid}.{version}";
            }
            catch
            {
                int version = -1;
                for (int i = 0; i < 20; i++)
                {
                    if (FileManagerSecure.PackageExists($"{packageUid}.{i}")) version = i;
                }
                if (version == -1) return null;
                return $"{packageUid}.{version}";
            }
            
        }
    }

    public static class MyExtensions
    {
        public static void Print(this object f)
        {
            SuperController.LogMessage(f.ToString());
        }
        public static void NullCheck(this object f)
        {
            SuperController.LogMessage((f == null).ToString());
        }
    }
}